function [eigval,eigvect,upperbound,flag] = PM_tensor(A,P,tol,opt,starting_point,D)
% MATRIX: PM_tensor computes the (p,q)-norm of a non-negative weakly 
% irreducible matrix A for 1 < p,q < inf so that 1 <= (p-1)*(q-1). 
% PM_tensor computes
%
%     max{|x'*A*y| s.t. ||x||_p = 1, ||y||_q = 1}
%
% where ||x||_p = norm(x,p) and ||y||_q = norm(y,q). Note that
% a matrix is a tensor of order 2 and the following explanations are
% directly made for the more general case of tensors.
%
% TENSOR: Given a non-negative weakly irreducible tensor A of order m and 
% 1 < p_1,...,p_m < inf such that there exists i with 
%             (d-1)(p_i/(p_i-1))<= p_k for all k =/= i
% PM_tensor computes the projective norm of A w.r.t. p_1,...,p_m, i.e.
%                   
%      max{|A(v_1,...,v_m)| s.t. ||v_i||_{p_i} = 1 for i = 1,...,m} 
%
% where:
% - A(v_1,...,v_m) is the multilinear form associated to the tensor A
%   evaluated at v_1 in R^{d_1}, v_2 in R^{d_2}, ... , v_m in R^{d_m}, i.e.
% A(x)=sum_{1<=j_1<=d_1,...,1<=j_m<=d_m}A_{j_1,...,j_m}*(v_1)_{j_1}*...*(v_m)_{j_m} 
%
% - ||v_i||_{p_i} is the p_i-norm on R^{d_i}, i.e. ||v_i||_{p_i}=norm(v_i,p_i)
%
% INPUT:
% - A = Tensor in matlab format (see example of use). For tensor input in
%       vector format see input D.
%
% - (OPTIONAL): P = row vector containing the constraints parameter
%                   p_1,...,p_m. Default value is p_1=...=p_m=m.
%
% - (OPTIONAL): tol = tolerance for the stopping criteria. Default value is
%                     1e-8. Note that the output singular value/vector
%                     are much more accurate than this tolerance. The
%                     algorithm stops after 1000 iterations or 
%                     |(L_+ - L_-)/L_+| < tol, where L_+ resp. L_- are
%                     upper and lower bounds on the norm of A computed
%                     at each iteration.
%
% - (OPTIONAL): opt = parameter for tracking convergence:
%                      opt = 0 -> Default value, normal run.
%                      opt = 1 -> Print the stopping criteria, the bounds 
%                                 on the norm and the objective at each 
%                                 iteration.
%                      opt = 2 -> Print the same as opt = 1 and additionally
%                                 plots a graphic showing ||x^k-x^*||_2,
%                                 |R(x^k)-R(x^*)| and the stopping criteria 
%                                 at each iteration (x^k is the current iterate 
%                                 and x^* the output).
%
% - (OPTIONAL): starting_point = choose the starting point for the
%                                Power-Method. Note that this vector should 
%                                be chosen with strictly positive entries.
%                                Default is ones(sum(size(A))). 
%
% - (OPTIONAL): D = Dimensions of the tensor. If you wish to give A as a
%                   vector containing the entries of the tensor in the
%                   reversed lexicographic order, precise its dimensions in
%                   the vector of integer D (having length equal to the
%                   order of the tensor).
%
% OUTPUT:
% - eigval = norm of A (a.k.a. l^{p_1,..,p_m} maximal singular value)
%
% - eigvect = maximizer (a.k.a. singular vector associated to eigval). This
%             vector (of size sum(size(A))) is the concatenation of the
%             maximizers x_1^*,..,x_m^*, i.e. eigvect = [x_1^* x_2^* ... x_m^*]
%             where x_i^* in R^{d_i}, ||x^*_i||_{p_i}=1 for i=1,...,m and
%                       norm of A = A(x_1^*,...,x_m^*)
%
% - upperbound = upper bound on the norm of A satisfying 
%                (1-tol)*upperbound <= norm of A <= upperbound
%
% - flag = convergence guarantee: 
%             flag = 1 -> output is guaranteed to be 
%                         the global maximizer and the upperbound 
%                         is guaranteed to satisfy the inequality above.
%             flag = 0 -> one of the assumption required is not fullfilled.
%
% REMARK: If you want to evaluate the objective at a given vector, use the
%         function Rayleigh_tensor. In particular, it holds
%                     Rayleigh_tensor(A,eigvect,P)=eigval
%         Type 'help Rayleigh_tensor' for more information.
%
% EXAMPLES OF USE:
%  - Ex. 1:
%  [eigval,eigvect,upperbound,flag]=PM_tensor(rand(2,3),[2,3])            
%
%  - Ex. 2:
%  [eigval,eigvect,upperbound,flag]=PM_tensor(rand(10,11,25),[2.5,8,3],1e-7)
%
%  - Ex. 3:
%  [eigval,eigvect,upperbound,flag]=PM_tensor(rand(3,3,3,3),[3.6,8,5,4],1e-7,0,1+rand(1,12))
%
%  - Ex. 4:
%  [eigval,eigvect,upperbound,flag]=PM_tensor(rand(1,2*3*5*8*2),[5.6,8.1,5.2,5.3,12.2],...
%                                             1e-7,2,ones(1,2+3+5+8+2),[2,3,5,8,2])

%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%

% Change here the maximum number of iteration

MAXIT = 1000;

%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%

% Filling missing arguments
%
if nargin < 3
    tol = 1e-8;
else
    if tol <= 0 | tol >=1
        error('Please choose a tolerance in the open interval (0,1)');
    end
end

if nargin < 6
    D=size(A);
end

n = sum(D);
d = length(D);

if nargin < 2
    P=d*ones(1,d);
end

Assumptions = ones(1,3); % =[non negaitivity, weak irreducibility, p_i assumption]
if nargin < 5
    starting_point = rand(1,n);
else
    if length(starting_point)~=sum(D)
        error('The chosen starting point does not have the correct dimension, i.e. length(starting_points) =/= sum(size(A))');
    end
end

if nargin < 4
    opt = 0;
end

if opt > 0
    if opt > 1
        X = [];
    end
    LMAX = [];
    LMIN = [];
    OBJ = [];
end

% Sanity checks
%

if min(A)<0
    warning('The tensor has negative entries   => there is no guarantee of convergence');
    Assumptions(1)=0;
end
if length(P)~=length(D)
    error('The vector P containing p_1,p_2,...,p_m should be the same size as the order of the tensor');
end
if min(P)<=1
    warning('One of the p_1,p_2,...,p_m is not > 1   => there is no guarantee of convergence');
    Assumptions(3)=-1;
end

% Assumptions check:
%

% Search p_i in {p_1,...,p_m} such that p_i satisfies the required assumption
if Assumptions(3)~=-1
    [pmin,Indmin] = min(P);
    [pmax,Indmax] = max(P);
    ppmin = pmin/(pmin-1);
    ppmax = pmax/(pmax-1);
    flagmin = 1;
    flagmax = 1;
    for i = 1:length(P)
        if i ~= Indmin
            if (d-1)*ppmin > P(i)
                flagmin = 0;
            end
        end
        if i ~= Indmax
            if (d-1)*ppmax > P(i)
                flagmax = 0;
            end
        end
    end
    if flagmin == 1
        Index = Indmin;
    else
        if flagmax == 1
            Index = Indmax;
        else
            warning('The assumption (d-1)(p_i/(p_i-1))<= p_k for all k =/= i isnt satisfied   => there is no guarantee of convergence');
            Index = 1;
            Assumptions(3)=0;
        end
    end
else
    Index = 1;
end
if Index == 1
    Imin = 1;
    Imax = D(1);
else
    Imin = sum(D(1:Index-1))+1;
    Imax = sum(D(1:Index));
end

% Weak irreducibility
bol = wc(A,D);
if bol == 0
    Assumptions(2)=0;
    warning('The tensor is not weakly irreducible   => there is no guarantee of convergence');
end


% The Power Method
%

% Initialization
xold = starting_point;
jmin = 1;
jmax = D(1);
for j = 1:d
    xold(jmin:jmax)=xold(jmin:jmax)/norm(xold(jmin:jmax),P(j));
    jmin = jmax+1;
    if j<d
        jmax = jmax + D(j+1);
    end
end
if opt > 0
    v = xold(Imin:Imax);
    xold(Imin:Imax) = single_supergrad(A,xold,D,Index-1);
    OBJ = [OBJ dot(v,xold(Imin:Imax))];
else
    xold(Imin:Imax) = single_supergrad(A,xold,D,Index-1);
end
xold(Imin:Imax) = (abs(xold(Imin:Imax)).^(P(Index)-1)).*sign(xold(Imin:Imax));
xnew = supergrad(A,xold,D);
ji_min = 1;
ji_max = D(1);
lmin = 1;
lmax = 1;
for i = 1:d
    if i~= Index
        xo_i = (abs(xold(ji_min:ji_max).^(P(i)-1))).*sign(xold(ji_min:ji_max));
        xn_i = xnew(ji_min:ji_max);
        lmax = lmax*max(xn_i(xo_i~=0)./xo_i(xo_i~=0));
        lmin = lmin*min(xn_i(xo_i~=0)./xo_i(xo_i~=0));
        xn_i = abs(xn_i).^(1/(P(i)-1)).*sign(xn_i);
        xold(ji_min:ji_max) = xn_i/norm(xn_i,P(i));
    end
    if i < length(D)
        ji_min = ji_min + D(i);
        ji_max = ji_max + D(i+1);
    end
end
lmax = lmax^((P(Index)-1)/(P(Index)*(d-1)));
lmin = lmin^((P(Index)-1)/(P(Index)*(d-1)));
iter=0;
if opt > 0
    if opt > 1
        X = [X ; xold];
    end
    LMAX = [LMAX lmax];
    LMIN = [LMIN lmin];
end
K = ((P(Index)-1)/(P(Index)*(d-1)));
if opt>0
    fprintf('L_+  = upper bound on the norm : L_-  = lower bound on the norm : Current estimate of the norm : Stopping criteria \n')
end
% Ierations
while abs((lmax-lmin)/lmax) > tol && iter < MAXIT
    iter = iter+1;
    if opt > 1
        v = xold(Imin:Imax)/norm(xold(Imin:Imax),P(Index));
        xold(Imin:Imax) = single_supergrad(A,xold,D,Index-1);
        OBJ = [OBJ dot(v,xold(Imin:Imax))];
    else
        xold(Imin:Imax) = single_supergrad(A,xold,D,Index-1);
    end
    xold(Imin:Imax) = (abs(xold(Imin:Imax)).^(1/(P(Index)-1))).*sign(xold(Imin:Imax));
    xnew = supergrad(A,xold,D);
    ji_min = 1;
    ji_max = D(1);
    lmin = 1;
    lmax = 1;
    for i = 1:d
        if i~= Index
            xo_i = (abs(xold(ji_min:ji_max).^(P(i)-1))).*sign(xold(ji_min:ji_max));
            xn_i = xnew(ji_min:ji_max);
            lmax = lmax*max(xn_i(xo_i~=0)./xo_i(xo_i~=0));
            lmin = lmin*min(xn_i(xo_i~=0)./xo_i(xo_i~=0));
            xn_i = abs(xn_i).^(1/(P(i)-1)).*sign(xn_i);
            xold(ji_min:ji_max) = xn_i/norm(xn_i,P(i));
        end
        if i < length(D)
            ji_min = ji_min + D(i);
            ji_max = ji_max + D(i+1);
        end
    end
    lmax = lmax^K;
    lmin = lmin^K;
    if opt>0
        fprintf('Iteration %i : L_+ = %.16f : L_- = %.16f :norm approx. =  %.16f : |(L_+ - L_-)/L_+| = %.16f \n',iter,lmax,lmin,OBJ(end),(lmax-lmin)/lmax);
    end
    if opt>0
        if opt > 1
            X=[X ; xold];
        end
        LMAX = [LMAX lmax];
        LMIN = [LMIN lmin];
    end
end
% Building output
upperbound = lmax;
eigvect = xold;
v = single_supergrad(A,xold,D,Index-1);
v = (abs(v).^(1/(P(Index)-1))).*sign(v);
v = v/norm(v,P(Index));
eigvect(Imin:Imax) = v;
eigval = dot(eigvect(1:D(1)),single_supergrad(A,eigvect,D,0));
if Assumptions == ones(1,3)
    flag = 1;
else
    flag = 0;
end


% Make Plots
%
if opt > 1
    nu = length(LMAX);
    DIFFVECT = zeros(1,nu-1);
    DIFFOBJ = zeros(1,nu-1);
    DIFFL = zeros(1,nu-1);
    for i = 1:nu
        DIFFL(i)=abs(LMAX(i)-LMIN(i))/LMAX(i);
        if i< nu
            DIFFVECT(i) = norm(X(end,:)-X(i,:));
            DIFFOBJ(i) = abs(OBJ(end)-OBJ(i));
        end
    end
    figure
    semilogy((1:nu-1),DIFFL(1:nu-1),'r','linewidth',2)
    hold on
    semilogy((1:nu-1),DIFFOBJ,'k','linewidth',2)
    hold on
    semilogy((1:nu-1),DIFFVECT,'b','linewidth',2)    
    TL = title('Convergence plots');
    set(TL,'Interpreter','latex');
    set(TL,'FontSize',14);
    LEGE=legend('\begin{tabular}{c}Stopping criteria: $\left|\frac{L_+ - L_-}{L^{\phantom{1}}_+}\right|$\end{tabular}\quad','\begin{tabular}{c}Iterates convergence:\\ $\|x^k - x^*\|_2$\end{tabular}','\begin{tabular}{c}Objective convergence:\\ $|R(x^k) - R(x^*)|$\end{tabular}');
    set(LEGE,'Interpreter','latex');
    set(LEGE,'FontSize',14);
    XLAB = xlabel('iteration k');
    set(XLAB,'Interpreter','latex');
    set(XLAB,'FontSize',14);
end
                